"""
Provides utilities for working with objects, defaults, etc. to abstract away common idoms
"""

__all__ = []
from .core import *; from .core import __all__ as _all
__all__ += _all
from .Options import *; from .Options import __all__ as _all
__all__ += _all


"""
Convenience functions that are inefficient, but are maybe a bit easier to work with?
"""

from .. import Iterators as itut
from .. import Numputils as nput

import itertools
from collections import namedtuple
import numpy as np

from .CoordinateSystems import *

__all__ = [
    "cartesian_to_zmatrix",
    "zmatrix_to_cartesian",
    "canonicalize_internal",
    "enumerate_zmatrices",
    "extract_zmatrix_internals"
]

def cartesian_to_zmatrix(coords, ordering=None, use_rad = True):
    """
    Converts Cartesians to Z-Matrix coords and returns the underlying arrays

    :param coords: input Cartesians
    :type coords: np.ndarray
    :param ordering: the Z-matrix ordering as a list of lists
    :type ordering:
    :return: Z-matrix coords
    :rtype: Iterable[np.ndarray | list]
    """

    zms = CoordinateSet(coords, system=CartesianCoordinates3D).convert(ZMatrixCoordinates, ordering=ordering, use_rad=use_rad)
    crds = namedtuple("zms", ["coords", "ordering"])
    return crds(np.asarray(zms), zms.converter_options['ordering'])

def zmatrix_to_cartesian(coords, ordering=None, origins=None, axes=None, use_rad=True):
    """
    Converts Z-maztrix coords to Cartesians

    :param coords:
    :type coords: np.ndarray
    :param ordering:
    :type ordering:
    :param origins:
    :type origins:
    :param axes:
    :type axes:
    :param use_rad:
    :type use_rad:
    :return:
    :rtype:
    """
    carts = CoordinateSet(coords, ZMatrixCoordinates).convert(CartesianCoordinates3D,
                                                            ordering=ordering, origins=origins, axes=axes, use_rad=use_rad)
    return np.asarray(carts)


def canonicalize_internal(coord):
    dupes = len(np.unique(coord)) < len(coord)
    if dupes: return None
    if len(coord) == 2:
        i, j = coord
        if i > j:
            coord = (j, i)
    elif len(coord) == 3:
        i, j, k = coord
        if i > k:
            coord = (k, j, i)
    elif len(coord) == 4:
        i, j, k, l = coord
        if i > l:
            coord = (l, k, j, i)
    elif coord[0] > coord[-1]:
        coord = tuple(reversed(coord))
    return coord

def _zmatrix_iterate(coords, natoms=None,
                     include_origins=False,
                     canonicalize=True,
                     deduplicate=True,
                     allow_completions=False
                     ):
    # TODO: this fixes an atom ordering, to change that up we'd need to permute the initial coords...
    if canonicalize:
        coords = [tuple(reversed(canonicalize_internal(s))) for s in coords]

    if deduplicate:
        dupes = set()
        _ = []
        for c in coords:
            if c in dupes: continue
            _.append(c)
            dupes.add(c)
        coords = _

    if include_origins:
        if (1, 0) not in coords:
            coords = [(1, 0)] + coords
        if (2, 1) not in coords and (2, 0) not in coords:
            if (2, 1, 0) in coords:
                coords = [(2, 1)] + coords
            else:
                coords = [(2, 0)] + coords
        if (2, 0) in coords and (2, 0, 1) not in coords: # can this happen?
            coords.append((2,1,0))

    if natoms is None:
        all_atoms = {i for s in coords for i in s}
        natoms = len(all_atoms)

    dihedrals = [k for k in coords if len(k) == 4]
    all_dihedrals = [
        (i, j, k, l)
        for (i, j, k, l) in dihedrals
        if i > j and i > k and i > l
    ]

    # need to iterate over all N-2 choices of dihedrals (in principle)...
    # should first reduce over consistent sets
    if not allow_completions:
        dihedrals = [
            (i,j,k,l) for i,j,k,l in dihedrals
            if (i,j) in coords and (i,j,k) in coords
            # if (
            #         any(x in coords or tuple(reversed(x)) in coords for x in [(i,j), (l,k)])
            #         and any(x in coords or tuple(reversed(x)) in coords for x in [(i,j,k), (l,k,j)])
            # )
        ]

    embedding = [
        x for x in [(2, 0, 1), (2, 1, 0)]
        if x in coords
    ]

    # we also will want to sample from dihedrals that provide individual atoms
    atom_diheds = [[] for _ in range(natoms)]
    for n,(i,j,k,l) in enumerate(dihedrals):
        atom_diheds[i].append((i,j,k,l))

    # completions = []
    # if allow_completions:
    #     for d in all_dihedrals:
    #         if d in dihedrals: continue
    #         completions.extend([d[:2], d[:3]])
    #
    #     c_set = set()
    #     for d in dihedrals:
    #         c_set.add(d[:2])
    #         c_set.add(d[:3])
    #     coord_pairs = [
    #         (c[:2],c[:3])
    #         for
    #     ]
    #     for d in all_dihedrals:
    #         if d in dihedrals: continue
    #         completions.extend([d[:2], d[:3]])

    for dihed_choice in itertools.product(embedding, *atom_diheds[3:]):
        emb, dis = dihed_choice[0], dihed_choice[1:]
        yield (
            (0, -1, -1, -1),
            (1, 0, -1, -1),
            emb + (-1,)
        ) + dis

def enumerate_zmatrices(coords, natoms=None,
                        allow_permutation=True,
                        include_origins=False,
                        canonicalize=True,
                        deduplicate=True,
                        preorder_atoms=True,
                        allow_completions=False
                        ):
    if canonicalize:
        coords = [tuple(reversed(canonicalize_internal(s))) for s in coords]

    if deduplicate:
        dupes = set()
        _ = []
        for c in coords:
            if c in dupes: continue
            _.append(c)
            dupes.add(c)
        coords = _

    if natoms is None:
        all_atoms = {i for s in coords for i in s}
        natoms = len(all_atoms)

    if preorder_atoms:
        counts = itut.counts(itertools.chain(*coords))
        max_order = list(sorted(range(natoms), key=lambda k:-counts[k]))
    else:
        max_order = np.arange(natoms)

    for atoms in (
            itertools.permutations(max_order)
                if allow_permutation else
            [max_order]
    ):
        atom_perm = np.argsort(atoms)
        perm_coords = [
            tuple(reversed(canonicalize_internal([atom_perm[c] for c in crd])))
            for crd in coords
        ]
        for zm in _zmatrix_iterate(perm_coords,
                                   natoms=natoms,
                                   include_origins=include_origins,
                                   canonicalize=False,
                                   deduplicate=False,
                                   allow_completions=allow_completions
                                   ):
            yield [
                [atoms[c] if c >= 0 else c for c in z]
                for z in zm
            ]

def extract_zmatrix_internals(zmat):
    specs = []
    for row in zmat:
        if row[1] < 0: continue
        specs.append(tuple(row[:2]))
        if row[2] < 0: continue
        specs.append(tuple(row[:3]))
        if row[3] < 0: continue
        specs.append(tuple(row[:4]))
    return specs




import itertools

import scipy.linalg

from .VectorOps import vec_normalize, vec_angles
from . import VectorOps as vec_ops
from . import Misc as misc
import math, numpy as np, scipy as sp

__all__ = [
    "rotation_matrix",
    "skew_symmetric_matrix",
    "rotation_matrix_skew",
    "youla_skew_decomp",
    "youla_skew_matrix",
    "youla_angles",
    "youla_matrix",
    "skew_from_rotation_matrix",
    "translation_matrix",
    "affine_matrix",
    "extract_rotation_angle_axis"
]

#######################################################################################################################
#
#                                                 rotation_matrix
#

def rotation_matrix_2d(theta):
    return np.moveaxis(
        np.array([
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]
        ]),
        0, -2
    )

def rotation_matrix_basic(xyz, theta):
    """rotation matrix about x, y, or z axis

    :param xyz: x, y, or z axis
    :type xyz: str
    :param theta: counter clockwise angle in radians
    :type theta: float
    """

    axis = xyz.lower()
    theta = np.asanyarray(theta)
    one = np.ones_like(theta)
    zero = np.zeros_like(theta)
    if axis == "z": # most common case so it comes first
        mat = [
            [ np.cos(theta), -np.sin(theta), zero],
            [ np.sin(theta),  np.cos(theta), zero],
            [ zero,                    zero,  one]
        ]
    elif axis == "y":
        mat = [
            [ np.cos(theta), zero, -np.sin(theta)],
            [ zero,           one,           zero],
            [ np.sin(theta), zero,  np.cos(theta)]
        ]
    elif axis == "x":
        mat = [
            [  one,           zero,           zero],
            [ zero,  np.cos(theta), -np.sin(theta)],
            [ zero,  np.sin(theta),  np.cos(theta)]
        ]
    else:
        raise Exception("{}: axis '{}' invalid".format('rotation_matrix_basic', xyz))
    return np.moveaxis(np.array(mat), 0, -1)

#thank you SE for the nice Euler-Rodrigues imp: https://stackoverflow.com/questions/6802577/rotation-of-3d-vector
def rotation_matrix_ER(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([
        [aa + bb - cc - dd, 2 * (bc + ad),     2 * (bd - ac)    ],
        [2 * (bc - ad),     aa + cc - bb - dd, 2 * (cd + ab)    ],
        [2 * (bd + ac),     2 * (cd - ab),     aa + dd - bb - cc]
    ])

def rotation_matrix_ER_vec(axes, thetas):
    """
    Vectorized version of basic ER
    """

    axes = vec_normalize(np.asanyarray(axes))
    thetas = np.asanyarray(thetas)
    # if len(axes.shape) == 1:
    #     axes = axes/np.linalg.norm(axes)
    #     axes = np.broadcast_to(axes, (len(thetas), 3))
    # else:
    #     axes = vec_normalize(axes)

    ax_shape = axes.shape[:-1]
    t_shape = thetas.shape
    axes = np.reshape(axes, (-1, 3))
    thetas = thetas.reshape(-1)
    if thetas.ndim == 0:
        base_shape = ax_shape
    elif axes.ndim == 1:
        base_shape = t_shape
    elif thetas.ndim != axes.ndim - 1:
        raise ValueError(f"can't broadcast axes and angles with shapes {ax_shape} and {t_shape}")
    else:
        base_shape = tuple(a if t == 1 else t for a,t in zip(ax_shape, t_shape))

    a = np.cos(thetas/2.0)
    b, c, d = np.moveaxis(-axes * np.reshape(np.sin(thetas / 2.0), (len(thetas), 1)), -1, 0)
    v = np.array([a, b, c, d])
    # triu_indices
    rows, cols = (
        np.array([0, 0, 0, 0, 1, 1, 1, 2, 2, 3]),
        np.array([0, 1, 2, 3, 1, 2, 3, 2, 3, 3])
    )
    aa, ab, ac, ad, bb, bc, bd, cc, cd, dd = v[rows] * v[cols]
    ## Uses half-angle formula to get compact form for Euler-Rodrigues
    # a^2 * I + [[ 0,    2ad, -2ac]   + [[b^2 - c^2 - d^2,               2bc,              2bd]
    #            [-2ad,    0,  2ab],     [             2bc, -b^2 + c^2 - d^2,              2cd]
    #            [ 2ac, -2ab,    0]]     [             2bd,              2cd, -b^2 - c^2 + d^2]]
    R = np.array([
        [aa + bb - cc - dd,      2 * (bc + ad),         2 * (bd - ac)],
        [    2 * (bc - ad),  aa - bb + cc - dd,         2 * (cd + ab)],
        [    2 * (bd + ac),      2 * (cd - ab),     aa - bb - cc + dd]
    ])
    R = np.moveaxis(R, -1, 0)

    return R.reshape(base_shape + (3, 3))

def rotation_matrix_align_vectors(vec1, vec2):
    vec1 = vec_ops.vec_normalize(vec1)
    vec2 = vec_ops.vec_normalize(vec2)
    s = vec1 + vec2
    i = vec_ops.identity_tensors(vec1.shape[:-1], vec1.shape[-1])
    inner = 1 + vec1[..., np.newaxis, :] @ vec2[..., :, np.newaxis]

    return i - s[..., :, np.newaxis] * (s[..., np.newaxis, :]/inner) + 2 * vec1[..., :, np.newaxis] * vec2[..., np.newaxis, :]

def rotation_matrix(axis, theta=None):
    """
    :param axis:
    :type axis:
    :param theta: angle to rotate by (or Euler angles)
    :type theta:
    :return:
    :rtype:
    """
    if theta is None:
        skew_vector = np.asanyarray(axis, dtype=float)
        if skew_vector.ndim == 0:
            skew_vector = skew_vector[np.newaxis]
        base_shape = skew_vector.shape[:-1]
        skew_vector = skew_vector.reshape(-1, skew_vector.shape[-1])
        rots = np.array([
            rotation_matrix_skew(v)
            for v in skew_vector
        ])
        return rots.reshape(base_shape + rots.shape[-2:])
    elif isinstance(axis, str):
        if axis.lower() == '2d':
            return rotation_matrix_2d(theta)

        theta = np.asanyarray(theta)
        if len(axis) == 1 and theta.ndim == 1 or theta.shape[-2] != 1:
            theta = np.expand_dims(theta, -2)
        rot_mat = None
        for x,t in zip(axis.lower(), np.moveaxis(theta, -2, 0)):
            rr = rotation_matrix_basic(x, t)
            if rot_mat is None:
                rot_mat = rr
            else:
                rot_mat = rr @ rot_mat
        return rot_mat
    else:
        axis = np.asanyarray(axis)
        theta = np.asanyarray(theta)

        if axis.shape == theta.shape:
            return rotation_matrix_align_vectors(axis, theta)
        elif axis.ndim == theta.ndim + 1 and axis.shape[-1] == 3:
            return rotation_matrix_ER_vec(axis, theta)
        else:
            # we have the vectors that get mixed and their mixing angles, we assume any fixed axis is the 0 element
            # I haven't vectorized this at all...
            base_shape = axis.shape[:-2]
            theta = theta.reshape(-1, theta.shape[-1])
            axis = axis.reshape(-1, axis.shape[-2:])

            rots = np.array([
                rotation_matrix_from_angles_vectors(l, T)
                for T, l in zip(axis, theta)
            ])

            return rots.reshape(base_shape + rots.shape[-2:])

def skew_symmetric_matrix(upper_tri):
    upper_tri = np.asanyarray(upper_tri)
    l = upper_tri.shape[-1]
    n = (1 + np.sqrt(1 + 8*l)) / 2
    if int(n) != n:
        raise ValueError(f"vector of shape {l} doesn't correspond to the upper triangle of a matrix")
    n = int(n)
    base_shape = upper_tri.shape[:-1]
    m = np.zeros(base_shape + (n, n))
    rows, cols = np.triu_indices_from(m, 1)
    m[..., rows, cols] = upper_tri
    m[..., cols, rows] = -upper_tri
    return m

def extract_rotation_angle_axis(rot_mat, normalize=True):
    rot_mat = np.asanyarray(rot_mat)
    if rot_mat.shape[-1] == 2:
        return np.arccos(rot_mat[..., 0, 0]), None
    elif rot_mat.shape[-1] == 3:
        ang = np.arccos((np.trace(rot_mat, axis1=-2, axis2=-1) - 1) / 2)
        skew = (rot_mat - np.moveaxis(rot_mat, -1, -2)) / 2

        rows, cols = (np.array([2, 0, 1]), np.array([1, 2, 0]))
        ax = skew[..., rows, cols]
        if normalize:
            ax = vec_normalize(ax)
        else:
            ax = ax
        return ang, ax
    else:
        base_shape = rot_mat.shape[:-2]
        rot_mat = np.reshape(rot_mat, (-1,) + rot_mat.shape[-2:])
        angles = []
        axes = []
        for r in rot_mat:
            U, Q = scipy.linalg.schur(r)
            angles.append(youla_angles(U))
            axes.append(Q)

        angles = np.array(angles)
        axes = np.array(axes)

        return angles.reshape(base_shape + angles.shape[-1:]), axes.reshape(base_shape + axes.shape[-2:])

def youla_skew_decomp(A):
    n = len(A)
    s, T = sp.linalg.schur(A)

    l = np.diag(s, 1)
    if n % 2 == 0:
        start = 0
        end = n
    else:  # manage padding for odd dimension
        if abs(l[0]) < 1e-7:
            start = 1
            end = n
        else:
            start = 0
            end = n - 1
    l = l[start:end-1:2]

    return youla_matrix(l, n, axis_pos=0 if start == 0 else n), T

def youla_skew_matrix(l, n, axis_pos=0):

    U = np.zeros((n, n))
    o = np.concatenate([  # even inds
        np.arange(0, axis_pos, 2),
        np.arange(axis_pos + 1, n, 2),
    ])
    e = np.concatenate([  # odd inds
        np.arange(1, axis_pos, 2),
        np.arange(axis_pos + 2, n, 2),
    ])

    U[o, e] = l
    U[e, o] = -l

    return U

def youla_matrix(angles, n, axis_pos=0):

    cos = np.cos(angles)
    sin = np.sin(angles)

    # build 2x2 block mat
    U = np.eye(n)
    if n % 2 == 1:
        o = np.concatenate([  # even inds
            np.arange(0, axis_pos, 2),
            np.arange(axis_pos+1, n, 2),
            ])
        e = np.concatenate([  # even inds
            np.arange(1, axis_pos, 2),
            np.arange(axis_pos+2, n, 2),
            ])
    else:
        o = np.arange(0, n, 2)
        e = np.arange(1, n, 2)

    U[o, o] = cos
    U[e, e] = cos
    U[o, e] = sin
    U[e, o] = -sin

    return U

def youla_angles(U, axis_pos=None):
    l = np.arccos(np.round(np.diag(U), 8))
    n = len(U)
    if axis_pos is None:
        if n % 2 == 0:
            axis_pos = -1
        else:  # manage padding for odd dimension
            axis_pos = np.where(abs(l) < 1e-7)[0]
            axis_pos = 0 if axis_pos[0] == 0 else axis_pos[-1]
        if axis_pos < 0:
            axis_pos = n + axis_pos

    return np.concatenate([
        l[0:axis_pos:2],
        l[axis_pos+1::2]
    ])

def rotation_matrix_skew(upper_tri, create_skew=True):
    upper_tri = np.asanyarray(upper_tri)
    if create_skew:
        if (
                upper_tri.ndim < 2
                or upper_tri.shape[-1] != upper_tri.shape[-2]
                or not np.allclose(upper_tri, -np.moveaxis(upper_tri, -2, -1))
        ):
            upper_tri = skew_symmetric_matrix(upper_tri)

    U, T = youla_skew_decomp(upper_tri)
    return T@U@T.T

def skew_from_rotation_matrix(rot_mat):
    U, Q = sp.linalg.schur(rot_mat)

    l = youla_angles(U)
    s = youla_skew_matrix(l, U.shape[0], axis_pos=0 if U[0, 0] == 1 else U.shape[0])
    A = Q @ s @ Q.T
    return A[np.triu_indices_from(A, k=1)]

def rotation_matrix_from_angles_vectors(l, T):
    n = T.shape[0]
    if n % 2 == 1 and len(l) == (n // 2) + 1: # the axis is encoded in l
        axis_pos = np.where(np.abs(l) > 2 * np.pi)[0]
        if len(axis_pos) == 0:
            axis_pos = np.where(np.abs(l) < 1e-7)[0]
            if len(axis_pos) == 0:
                raise ValueError(f"can't find fixed axis position from angle encoding {l}")
        axis_pos = axis_pos[0]
    else:
        axis_pos = 0
    return T @ youla_matrix(l, n, axis_pos=axis_pos) @ T.T

#######################################################################################################################
#
#                                                 translation_matrix
#

def translation_matrix(shift):
    share = np.asarray(shift)
    if len(share.shape) == 1:
        ss = share
        zs = 0.
        os = 1.
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        )
    else:
        zs = np.zeros((share.shape[0],))
        os = np.ones((share.shape[0],))
        ss = share.T
        mat = np.array(
            [
                [os, zs, zs, ss[0]],
                [zs, os, zs, ss[1]],
                [zs, zs, os, ss[2]],
                [zs, zs, zs, os   ]
            ]
        ).T
    return mat

#######################################################################################################################
#
#                                                 affine_matrix
#

def affine_matrix(tmat, shift):
    """Creates an affine transformation matrix from a 3x3 transformation matrix or set of matrices and a shift or set of vecs

    :param tmat: base transformation matrices
    :type tmat: np.ndarray
    :param shift:
    :type shift:
    :return:
    :rtype:
    """

    base_mat = np.asanyarray(tmat)
    if shift is None:
        return base_mat

    if base_mat.ndim > 2:
        shifts = np.asanyarray(shift)
        if shifts.ndim == 1:
            shifts = np.broadcast_to(shifts, (1,)*(base_mat.ndim-2) + shifts.shape)
        shifts = np.broadcast_to(shifts, base_mat.shape[:-2] + (3,))
        shifts = np.expand_dims(shifts, -1)
        mat = np.concatenate([base_mat, shifts], axis=-1)
        padding = np.array([0., 0., 0., 1.])
        padding = np.broadcast_to(
            np.broadcast_to(padding, (1,)*(base_mat.ndim-2) + padding.shape),
            mat.shape[:-2] + (4,)
        )
        padding = np.expand_dims(padding, -2)
        mat = np.concatenate([mat, padding], axis=-2)
    else:
        mat = np.concatenate([base_mat, shift[:, np.newaxis]], axis=-1)
        mat = np.concatenate([mat, [[0., 0., 0., 1.]]], axis=-2)
    return mat



import itertools, collections
import scipy.sparse as sparse, numpy as np
from .. import Numputils as nput
from .. import Iterators as itut

__all__ = [
    "EdgeGraph"
]

class EdgeGraph:
    __slots__ = ["labels", "edges", "graph", "map"]
    def __init__(self, labels, edges, graph=None, edge_map=None):
        self.labels = labels
        self.edges = np.asanyarray(edges)
        if graph is None:
            graph = self.adj_mat(len(labels), self.edges)
        self.graph = graph
        if edge_map is None:
            edge_map = self.build_edge_map(self.edges)
        self.map = edge_map

    @classmethod
    def adj_mat(cls, num_nodes, edges):
        adj = np.zeros((num_nodes, num_nodes), dtype=bool)
        rows,cols = edges.T
        adj[rows, cols] = 1
        adj[cols, rows] = 1

        return sparse.csr_matrix(adj)

    @classmethod
    def build_edge_map(cls, edge_list):
        map = {}
        for e1,e2 in edge_list:
            if e1 not in map: map[e1] = set()
            map[e1].add(e2)
            if e2 not in map: map[e2] = set()
            map[e2].add(e1)
        return map

    @classmethod
    def _remap(cls, labels, pos, rows, cols):
        if len(rows) == 0:
            edge_list = np.array([], dtype=int).reshape(-1, 2)
        else:
            new_mapping = np.zeros(len(labels), dtype=int)
            new_mapping[pos,] = np.arange(len(pos))
            new_row = new_mapping[rows,]
            new_col = new_mapping[cols,]
            edge_list = np.array([new_row, new_col]).T

        return [labels[p] for p in pos], edge_list
    @classmethod
    def _take(cls, pos, labels, adj_mat:sparse.compressed):
        rows, cols, _ = sparse.find(adj_mat)
        utri = cols >= rows
        rows = rows[utri]
        cols = cols[utri]
        row_cont, _, _ = nput.contained(rows, pos)
        col_cont, _, _ = nput.contained(cols, pos)
        cont = np.logical_and(row_cont, col_cont)

        labels, edge_list = cls._remap(labels, pos, rows[cont], cols[cont])
        return cls(labels, edge_list)

    def take(self, pos):
        return self._take(pos, self.labels, self.graph)

    def split(self, backbone_pos):
        new_adj = self.graph.copy()
        for n in backbone_pos:
            for i,j in self.map[n]:
                new_adj[i,j] = 0
        ncomp, labels = sparse.csgraph.connected_components(new_adj, directed=False, return_labels=True)
        groups, _ = nput.group_by(np.arange(len(labels)), labels)
        return [
            self._take(pos, self.labels, new_adj)
            for _, pos in groups
        ]

    @classmethod
    def _bfs(cls, test, root, edge_map, visited:set):
        if root in visited:
            return
        queue = collections.deque([root])
        while queue:
            head = queue.pop()
            visited.add(head)
            test(head)
            queue.extend(h for h in edge_map.get(head, []) if h not in visited)
    @classmethod
    def _subgraph_match(cls,
                        root1, labels1, edge_map1,
                        root2, labels2, edge_map2,
                        visited=None
                        ):
        if labels1[root1] != labels2[root2]:
            return False
        if len(edge_map1.get(root1, [])) != len(edge_map2.get(root2, [])):
            return False

        if visited is None:
            visited = (set(), set())
        visit1, visit2 = visited

        # queue1 = collections.deque([root1])
        tests1 = set(edge_map1.get(root1, [])) - visit1
        tests2 = set(edge_map2.get(root2, [])) - visit2
        for r1 in tests1:
            visit1.add(r1)
            for r2 in tests2:
                visit2.add(r2)
                # check if all the subgraphs match via DFS
                # we could do this without recursion, but it'd be
                # more annoying
                if cls._subgraph_match(
                    r1, labels1, edge_map1,
                    r2, labels2, edge_map2,
                    visited=(visit1, visit2)
                ):
                    tests2.remove(r2)
                    break
                else:
                    visit2.remove(r2)

        return len(tests2) == 0

    @classmethod
    def graph_match(cls, graph1:'EdgeGraph', graph2:'EdgeGraph'):
        # we do some quick prunes
        atoms1 = graph1.labels
        atoms2 = graph2.labels
        if (
                len(atoms1) != len(atoms2)
                or atoms1[0] != atoms2[0]
                or len(graph1.edges) != len(graph2.edges)
                or list(sorted(atoms1)) != list(sorted(atoms2))
                or list(sorted(len(v) for v in graph1.map.values())) != list(sorted(len(v) for v in graph2.map.values()))
        ):
            return False

        return cls._subgraph_match(
            0, graph1.labels, graph1.map,
            0, graph2.labels, graph2.map
        )

    def __eq__(self, other):
        return self.graph_match(self, other)

    @classmethod
    def build_neighborhood_graph(cls, node, labels, edge_map, ignored=None, num=1):
        edges = []
        visited = set()
        if ignored is None: ignored = []
        ignored = set(ignored)
        queue = [node]
        for i in range(num):
            new_queue = []
            for node in queue:
                visited.add(node)
                new_nodes = set(edge_map[node]) - visited - ignored
                edges.extend((node, e) for e in new_nodes)
                new_queue.extend(new_nodes)
            queue = new_queue

        edges = np.array(edges, dtype=int)
        if len(edges) == 0:
            edges = np.reshape(edges, (-1, 2))
        labels, edges = cls._remap(labels, list(visited), edges[:, 0], edges[:, 1])
        return cls(labels, edges)

    def neighbor_graph(self, root, ignored=None, num=1):
        return self.build_neighborhood_graph(root, self.labels, self.map, ignored=ignored, num=num)

    def get_rings(self):
        # use rdkit's cycles...not assured to be the smallest set ;_;
        from ..ExternalPrograms.RDKit import RDMolecule
        return RDMolecule.from_coords(
            ["C"]*len(self.labels),
            coords=np.zeros((len(self.labels), 3)),
            bonds=[[int(i), int(j), 1] for i,j in self.edges]
        ).rings

    def get_fragments(self):
        from scipy.sparse import csgraph
        _, labels = csgraph.connected_components(self.graph, directed=False, return_labels=True)
        _, groups = nput.group_by(np.arange(len(labels)), labels)[0]
        return groups

    @classmethod
    def get_maximum_overlap_permutation(cls, graph_1:'EdgeGraph', graph_2:'EdgeGraph'):
        syms_1 = graph_1.labels
        syms_2 = graph_2.labels

        if any(s_1 != s_2 for s_1, s_2 in zip(syms_1, syms_2)):
            if len(itut.dict_diff(itut.counts(syms_1), itut.counts(syms_2))) > 0:
                raise ValueError(f"graph labels must agree: {syms_1} != {syms_2}")
            ordering_1 = list(sorted(range(len(syms_1)), key=syms_1.__getitem__))
            ordering_2 = list(sorted(range(len(syms_2)), key=syms_2.__getitem__))
            perm_0 = np.array(ordering_2)[np.argsort(ordering_1)]
            graph_2 = graph_2.take(perm_0)
        else:
            perm_0 = None

        bond_set_1 = {tuple(sorted(e)) for e in graph_1.edges}
        bond_set_2 = {tuple(sorted(e)) for e in graph_2.edges}

        # initial bond difference
        test_bonds = np.unique(np.array(
            list(bond_set_1 - bond_set_2)
            + list(bond_set_2 - bond_set_1)
        ))

        # permutable groups
        sym_splits, _ = nput.group_by(np.arange(len(syms_1)), np.array([ord(s) for s in syms_1]))
        perm_blocks = []
        perm_atoms = []
        for _, atom_inds in zip(*sym_splits):
            atom_inds = nput.intersection(atom_inds, test_bonds)[0]  # only permute things in the original core
            if len(atom_inds) > 0:
                perm_atoms.append(atom_inds)
                perm_blocks.append(itertools.permutations(atom_inds))

        nsym = len(syms_1)
        core_size = len(test_bonds)
        perm = np.arange(nsym)
        for full_perm in itertools.product(*perm_blocks):
            reindexing = np.arange(nsym)
            for atom_inds, new_idx in zip(perm_atoms, full_perm):
                reindexing[atom_inds,] = new_idx
            new_bond_set_1 = {
                (reindexing[i], reindexing[j])
                for (i, j) in bond_set_1
            }
            new_core = np.unique(np.array(
                list(new_bond_set_1 - bond_set_2)
                + list(bond_set_2 - new_bond_set_1)
            ))
            if len(new_core) < core_size:
                perm = reindexing

        if perm_0 is not None:
            return perm_0[perm]
        else:
            return perm

    def get_reindexing(self, other_graph):
        return self.get_maximum_overlap_permutation(other_graph, self)
    def align_labels(self, other_graph):
        perm = self.get_reindexing(other_graph)
        return self.take(perm)


"""
Provides a set of singleton objects that can declare their purpose a little bit better than None can
"""

# import enum
import types

__all__ = [
    "default",
    "is_default",
    "handle_default",
    "uninitialized",
    "is_uninitialized",
    "handle_uninitialized",
    "is_interface_like",
    "is_dict_like",
    "cached_eval"
]

class SingletonType:
    """
    A base type for singletons
    """
    __slots__ = []

class DefaultType(SingletonType):
    """
    A type for declaring an argument should use its default value (for when `None` has meaning)
    """
    __is_default__ = True
default=DefaultType()

class UninitializedType(SingletonType):
    """
    A type for declaring an argument should use its default value (for when `None` has meaning)
    """
    __is_uninitialized__ = True
uninitialized = UninitializedType()

def is_interface_like(obj, interface_types, implementation_attrs):
    return (
        isinstance(obj, interface_types)
        or all(hasattr(obj, a) for a in implementation_attrs)
    )

def is_dict_like(obj,
                 interface_types=(dict, types.MappingProxyType),
                 implementation_props=('items',)
                 ):
    return is_interface_like(obj, interface_types, implementation_props)

def is_default(obj, allow_None=True):
    if allow_None and obj is None:
        return True

    return (
            obj is default
            or isinstance(obj, DefaultType)
            or (hasattr(obj, '__is_default__') and obj.__is_default__)
    )

def handle_default(opt, default_value, allow_None=True):
    if is_default(opt, allow_None=allow_None):
        return default_value
    else:
        return opt

def is_uninitialized(obj, allow_None=True):
    if allow_None and obj is None:
        return True

    return (
            obj is uninitialized
            or isinstance(obj, UninitializedType)
            or (hasattr(obj, '__is_uninitialized__') and obj.__is_uninitialized__)
    )

def handle_uninitialized(opt, initializer, allow_None=True, args=(), kwargs=None):
    if is_uninitialized(opt, allow_None=allow_None):
        return initializer(*args, **({} if kwargs is None else kwargs))
    else:
        return opt


def cached_eval(cache, key, generator, *,
                condition=None,
                args=(),
                kwargs=None):
    condition = (condition is None or condition(key))
    if not condition:
        if kwargs is None: kwargs = {}
        return generator(*args, **kwargs)

    if key in cache:
        return cache[key]

    if kwargs is None: kwargs = {}
    val = generator(*args, **kwargs)
    cache[key] = val

    return val